@@ -5,8 +5,9 @@ use swc_atoms::{js_word, JsWord};
 use swc_common::{
     ast_node, util::take::Take, EqIgnoreSpan, Span, Spanned, SyntaxContext, DUMMY_SP,
 };
+use unicode_xid::UnicodeXID;
 
-/// Identifer used as a pattern.
+/// Identifier used as a pattern.
 #[derive(Spanned, Clone, Debug, PartialEq, Eq, Hash, EqIgnoreSpan, Serialize, Deserialize)]
 #[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 pub struct BindingIdent {
@@ -85,6 +86,77 @@ pub struct Ident {
     pub optional: bool,
 }
 
+impl Ident {
+    /// Returns true if `c` is a valid character for an identifier start.
+    pub fn is_valid_start(c: char) -> bool {
+        c == '$' || c == '_' || c.is_ascii_alphabetic() || {
+            if c.is_ascii() {
+                false
+            } else {
+                UnicodeXID::is_xid_start(c)
+            }
+        }
+    }
+
+    /// Returns true if `c` is a valid character for an identifier part after
+    /// start.
+    pub fn is_valid_continue(c: char) -> bool {
+        c == '$' || c == '_' || c == '\u{200c}' || c == '\u{200d}' || c.is_ascii_alphanumeric() || {
+            if c.is_ascii() {
+                false
+            } else {
+                UnicodeXID::is_xid_continue(c)
+            }
+        }
+    }
+
+    /// Alternative for `toIdentifier` of babel.
+    ///
+    /// Returns [Ok] if it's a valid identifier and [Err] if it's not valid.
+    /// The returned [Err] contains the valid symbol.
+    pub fn verify_symbol(s: &str) -> Result<(), String> {
+        if s.is_reserved() || s.is_reserved_in_strict_mode(true) || s.is_reserved_in_strict_bind() {
+            let mut buf = String::with_capacity(s.len() + 1);
+            buf.push('_');
+            buf.push_str(s);
+            return Err(buf);
+        }
+
+        {
+            let mut chars = s.chars();
+
+            if let Some(first) = chars.next() {
+                if Self::is_valid_start(first) {
+                    if chars.all(Self::is_valid_continue) {
+                        return Ok(());
+                    }
+                }
+            }
+        }
+
+        let mut buf = String::with_capacity(s.len() + 2);
+        let mut has_start = false;
+
+        for c in s.chars() {
+            if !has_start && Self::is_valid_start(c) {
+                has_start = true;
+                buf.push(c);
+                continue;
+            }
+
+            if Self::is_valid_continue(c) {
+                buf.push(c);
+            }
+        }
+
+        if buf.is_empty() {
+            buf.push('_');
+        }
+
+        Err(buf)
+    }
+}
+
 /// See [Ident] for documentation.
 pub type Id = (JsWord, SyntaxContext);
 
